@@ -225,7 +225,7 @@ ngx_create_hashed_filename(ngx_path_t *path, u_char *file, size_t len)
 
     file[path->name.len + path->len]  = '/';
 
-    for (n = 0; n < 3; n++) {
+    for (n = 0; n < NGX_MAX_PATH_LEVEL; n++) {
         level = path->level[n];
 
         if (level == 0) {
@@ -249,7 +249,7 @@ ngx_create_path(ngx_file_t *file, ngx_path_t *path)
 
     pos = path->name.len;
 
-    for (i = 0; i < 3; i++) {
+    for (i = 0; i < NGX_MAX_PATH_LEVEL; i++) {
         if (path->level[i] == 0) {
             break;
         }
@@ -399,6 +399,8 @@ char *
 ngx_conf_merge_path_value(ngx_conf_t *cf, ngx_path_t **path, ngx_path_t *prev,
     ngx_path_init_t *init)
 {
+    ngx_uint_t  i;
+
     if (*path) {
         return NGX_CONF_OK;
     }
@@ -419,13 +421,10 @@ ngx_conf_merge_path_value(ngx_conf_t *cf, ngx_path_t **path, ngx_path_t *prev,
         return NGX_CONF_ERROR;
     }
 
-    (*path)->level[0] = init->level[0];
-    (*path)->level[1] = init->level[1];
-    (*path)->level[2] = init->level[2];
-
-    (*path)->len = init->level[0] + (init->level[0] ? 1 : 0)
-                   + init->level[1] + (init->level[1] ? 1 : 0)
-                   + init->level[2] + (init->level[2] ? 1 : 0);
+    for (i = 0; i < NGX_MAX_PATH_LEVEL; i++) {
+        (*path)->level[i] = init->level[i];
+        (*path)->len += init->level[i] + (init->level[i] ? 1 : 0);
+    }
 
     if (ngx_add_path(cf, path) != NGX_OK) {
         return NGX_CONF_ERROR;
@@ -518,7 +517,7 @@ ngx_add_path(ngx_conf_t *cf, ngx_path_t **slot)
                 return NGX_ERROR;
             }
 
-            for (n = 0; n < 3; n++) {
+            for (n = 0; n < NGX_MAX_PATH_LEVEL; n++) {
                 if (p[i]->level[n] != path->level[n]) {
                     if (path->conf_file == NULL) {
                         if (p[i]->conf_file == NULL) {
